//----------------------------------------------------------------------------------------
// Variables
//----------------------------------------------------------------------------------------
        .const zpa =                $10
        .const zpx =                $11
        .const zpy =                $12
        
        .const ACTIVESPRITEMASK =   %11111111
//----------------------------------------------------------------------------------------
// Macros
//----------------------------------------------------------------------------------------
.pseudocommand asr bitmask {alr bitmask}
//----------------------------------------------------------------------------------------
.macro delay(cyc) {
    ldy #floor((cyc-20)/5)
    jsr delay + 4 - mod(cyc-20, 5)
}       
//----------------------------------------------------------------------------------------
// Basic Upstart Macro
//----------------------------------------------------------------------------------------
        *= $0801 "Basic SYS-line"
        BasicUpstart(init)
//----------------------------------------------------------------------------------------
// Init
//----------------------------------------------------------------------------------------
        *=$0900 "Init"
init:   
        lda #$00
        sta $d015
        sta $d01a
        sta $dc0e
        sta $dc0f
        sta $dd0e
        sta $dd0f
        
        lda $dc0d		
        lda $dd0d
        
        lda #$7f
        sta $dc0d
        sta $dd0d
        
        lda #$35
        sta $01
        
        jsr set_timer
        
        lda #$00
        sta $d020
        lda #$06
        sta $d021
        lda #$08
        sta $d016
        lda #$f3
        sta $dd00
        lda #$35
        sta $d018
		lda #$18
		sta $d011
        
        ldy #$07
sprcol_pos_init_loop:
        lda #$03
        sta $d027,y
        lda spr_xpos_tab,y
        sta $d000,y
        lda spr_xpos_tab+8,y
        sta $d008,y
        dey
        bpl sprcol_pos_init_loop
        lda #$80
        sta $d010
        lda #ACTIVESPRITEMASK
        sta $d015
        sta $d01b
        
        lda #$00
        sta $d017
        sta $d01c
        sta $d01d
        
        lda #$20
        sta ld_prev_timerbyte+1
        lda #<irq
        sta $fffe
        lda #>irq
        sta $ffff
        
        lda $d011   // checks raster and cycle position after init; should give fixed
        sta $0cf0   // values if timer init etc. finished correctly
        lda $d012   //
        sta $0cf1   // for pal, we have $0cf0..2 = $18 $0c $1d
        lda $dc04   // for ntsc, it's   $0cf0..2 = $18 $0c $29
        sta $0cf2
        
		lda #$2f
		sta $d012
        sta $d019
		inc $d01a
//----------------------------------------------------------------------------------------
// Main
//----------------------------------------------------------------------------------------
// the opcodes in the "main thread" are chosen to represent
// all possible cycle lengths (of "legal" opcodes)
// this way the dejitter-routine gets to handle all 8 latency-states
//----------------------------------------------------------------------------------------
main:   
		ldx #$00
		ldy #$00
		inc $0cf3,x // $0cf3 used as "alive counter" for main thread
		bne *-3
loopie: 
        lda $0cf3
        sta $0cf3,x
        lda ($20,x)
		inc $0cf3,x
		iny
		bne loopie
		jmp main
//----------------------------------------------------------------------------------------
        *= $0a00
delay:  
        cpy #$c0
        cpy #$24
        nop
        dey
        bpl *-1
        rts
//----------------------------------------------------------------------------------------
set_timer:
chk_ntsc:
        ldx #TAX
        lda $d012
        bne chk_ntsc+1
        txa
        bmi chk_ntsc
        
        lda #$3e
        cpx #$06
        bne skip_ntsc_timing
        lda #$40
skip_ntsc_timing:
        sta $dc04
        and #$02
        sta call_delay_sub+1
        
        ldy #$06
        ldx #$00
        stx $dc05
        cpx $d012
        beq *-3
var_kill:
        inx
        lda #$11
call_delay_sub:
        jsr delay
        ldy #$05
        cpx $d012
        //---------------//
        beq *+2
        cpx #$06    // x-val is not yet incremented at this point, so cpx #$06 checks
                    // if line 7 is reached
		            // After leaving line 0 with a conituous [cpx $d012 : beq *-3]-loop,
		            // the actual variance elemination starts at cycle 4..10 in line 1
		            // the routine now egalises one cycle per line, giving a stable
		            // cycle postion as stated in the following table
		            // 
		            // start pos in line 1:  4 5 6 7 8 9 10
		            // stable pos in line:   7 6 5 4 3 2  2
		            //
		            // thus, the variance elemination loop ends at cycle 10 in line 7
branch: 
        bne var_kill
        :delay(23)
        sta $dc0e
        rts
//----------------------------------------------------------------------------------------
		*= * "IRQ"
//----------------------------------------------------------------------------------------
irq:	
		sta zpa
		stx zpx
		ldx #$7e
        
        lda $dc04
        
        sty zpy
        
        lsr
        bcs skip1
skip1:  
        asr #$03
        bcc skip2
        bcs skip2
skip4:  
        bne endvarkill
skip2:  
        bne skip4		
//--- rasline $2f @cyc#43
endvarkill:
        stx $d020
//--- rasline $2f @cyc#47
skipjmp:
        jmp skipmax
        // the jmp-adress is decreased in every frame so the [LDA timer] is shifted
        // to later cycle positions, at the rate of +1 cycle/frame
        // max. delay is 127 cycles; the next frame then restarts with zero delay
        //
        // the timer value uniquely defines the cycle position at which
        // the read acces of the [LDA timer] (=the last R-cycle) happened 
//----------------------------------------------------------------------------------------
spr_xpos_tab:
        .byte $18, $2f, $38, $2f, $58, $2f, $78, $2f
        .byte $98, $2f, $b8, $2f, $d8, $2f, $08, $2f
//----------------------------------------------------------------------------------------
        *= $0b00
//----------------------------------------------------------------------------------------
skipmin:
        .fill 125, $c9
        cmp #$c5
//--- rasline $2f @cyc#50 or later, depending on the entry point determined by skipjmp
skipmax:
        nop
        lda $dc04
        // in the following we assume no additional delay (=skipjmp to skipmax) and
        // that all 8 sprites are active
        // The four (R-)cycles of the [LDA timer] are denoted by R0, R1, R2 and R3, i.e.
        // cycle-wise we have [LDA timer] = [R0 R1 R2 R3]
        //
        // for pal, the R-cycles should occur as shown in the following table
        // cycle in line $2f:                    ...|52|53|54|55|56|57|58|59|60|61|62|63
        // R-cycles:                                 R0|R1|R2|(xxxxxx)|(sprite  fetches)
        // cycle in line $30:  1|2|3|4|5|6|7|8|9|10|11|12|13|14|15|...
        // R-cycles:           (..sprite fetches..)|R3|(xxxxxx)|(badline fetches)
        // 
        // the timer is configured as usual for C64 with PAL-VIC: continuous cycle count 
        // with init value $3e (reset every line at cycle 37)
        // this configuration gives timer-value = $19 at cycle position 11
        //
        // -> for PAL systems, the read value is $19, so it matches the theory
        //
        // with line $30 being a badline, the timer value will be the same (=$19) when
        // the [LDA timer] is delayed by one cycle, since the R3-cycle occurs at cycle 11
        // in line $31. Delaying the LDA-opcode further gives descending timer values.
        
        // for ntsc (65 cycle versions!), the R-cycles should be distributed as follows
        // (according to THE Vic-article):
        // cycle in line $2f:               ...|52|53|54|55|56|57|58|59|60|61|62|63|64|65
        // R-cycles:                            R0|R1|R2|R3|  |(xxxxxx)|(sprite  fetches)
        // cycle in line $30:  1|2|3|4|5|6|7|8|9|10|11|12|13|14|15|...
        // R-cycles:           (..sprite fetches..)|  |(xxxxxx)|(badline fetches)
        // 
        // here, cycle positions 56 in line $2f and 11 in line $30 are still "free"
        //
        // the timer is configured to suit ntsc-needs: continuous cycle count 
        // with init value $40 (also reset every line at cycle 37)
        // this configuration should give timer-value = $2e at cycle position 55
        // 
        // -> running the timercompare-prg on ntsc correctly gives timer-val = $2e
        //
        // BUT...
        //
        // delaying the [LDA timer] by one cycle, we should get timer-value = $2d
        // as cycle#56 in line $2f is free (and not belonging to the VIC-overtake-cycles
        // that follow directly after).
        //
        // ->-> suprisingly, the timercompare-prg gives timer-val = $1a instead!
        
        // Delaying furhter, one can i.e. test which cycle positions can be reached in 
        // line = $30. Since it is a badline, it should end like this:
        // cycle in line $30:               ...|52|53|54|55|56|57|58|59|60|61|62|63|64|65
        // R-cycles:                   (badline fetches)|R3|  |(xxxxxx)|(sprite  fetches)
        //
        // it is possible to move the R3-cycle to cycle 55 of line $30, at which the
        // timer value = $2e is read back (same case as described above)
        // to the opposite, it is *not* possible to move the R3-cycle to cycle 56 of
        // line $30!!!
        //
        // The timer values suggest that on NTSC-machines, we do not have TWO additional
        // "free" cycles before the VIC-overtake-cycles before the sprite-fetches start
        // but only one.
        //
        // for ntsc (65 cycle versions!), the VIC-cycles seem to be distributed in the 
        // following way (example shows badline with all sprites active):
        // cycle-pos:          1|2|3|4|5|6|7|8|9|10|11|12|13|14|15|...
        // VIC-function:       (sprite  fetches)|  |  |(xxxxxx)|(badline fetches)
        // cycle in line $2f:                ...|52|53|54|55|56|57|58|59|60|61|62|63|64|65
        // R-cycles:                    (badline fetches)|  |(xxxxxx)|(..sprite fetches..)
        //
        ldy #$00
        sty $d020
ld_prev_timerbyte:
        ldx #$20
cc0D_restore:
        stx $0d7f
        sta ld_prev_timerbyte+1
        ora #$80
cc0D:   
        sta $0d00
        
        lda cc0D+1
        sta cc0D_restore+1
        
        ldx skipjmp+1
        bne decjmpadr
        ldx #(<skipmax)+1
decjmpadr:
        dex
        stx skipjmp+1
        txa
        eor #$7f
        sta cc0D+1
//----------------------------------------------------------------------------------------
irq_end:
		ldy zpy		//zpy = $12
		ldx zpx		//zpx = $11
		lda zpa		//zpa = $10
		dec $d019
nmi:	rti
//----------------------------------------------------------------------------------------
//      Sprite Data
//----------------------------------------------------------------------------------------
        *=$0bc0
        .fill 3*8, $ff
        .fill 3*(21-8), $00 
        .byte $00
//----------------------------------------------------------------------------------------
//      Screen on Init
//----------------------------------------------------------------------------------------
        *=$0c00
        .fill 40, $2e
        .fill 40, $30+mod(i, 10)
        .fill 40, $30+floor(i/10) 
        .fill 22*40, $20
//----------------------------------------------------------------------------------------
//      Sprite ptr
//----------------------------------------------------------------------------------------
        *=$0ff8
        .fill 8, mod($0bc0, $4000)/$40
//----------------------------------------------------------------------------------------
